/*
 * copyright.c:
 *	include a copyright notice in the irc binaries.
 */

static char copyright[] =
"\
   IRC - Internet Relay Chat                                            \n\
   Based on Internet Relay Chat v2.0 by Jarkko Oikarinen		\n\
   Copyright (C) 1990, 1991 by Jarkko Oikarinen                         \n\
                                                                        \n\
   This program is free software; you can redistribute it and/or modify \n\
   it under the terms of the GNU General Public License as published by \n\
   the Free Software Foundation; either version 1, or (at your option)  \n\
   any later version.                                                   \n\
                                                                        \n\
   This program is distributed in the hope that it will be useful,      \n\
   but WITHOUT ANY WARRANTY; without even the implied warranty of       \n\
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        \n\
   GNU General Public License for more details.                         \n\
                                                                        \n\
   You should have received a copy of the GNU General Public License    \n\
   along with this program; if not, write to the Free Software          \n\
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            \n\
									\n\
   IRC is free software and you are welcome to distribute copies of it	\n\
   under certain conditions.  Consult the file COPYING, distributed in  \n\
   the source for IRC to see the conditions";
